import json

import names
import numpy as np
import pandas as pd
from randomtimestamp import randomtimestamp

from teaching_tools.ab_test.params import cohort_params


def load_json(filepath: str) -> dict:
    """Loads JSON file.

    Parameters
    ----------
    filepath : str
        Location of JSON file

    Returns
    -------
    dict

    """
    with open(filepath, "r") as f:
        return json.load(f)


def load_cohort_params(c) -> dict:
    """Helper function for loading cohort parameters.

    Parameters
    ----------
    c : Union[str, dict]
        If ``str``, treated as a filepath. If ``dict``, returns same.

    Returns
    -------
    dict

    """
    if isinstance(c, str):
        return load_json(c)
    elif isinstance(c, dict):
        return c
    else:
        return cohort_params


class StudentGenerator:
    def __init__(
        self,
        cohort_params=None,
        days=1,
        start=None,
        n=None,
        is_experiment=False,
    ):
        self.cohort_params = load_cohort_params(cohort_params)
        self.is_experiment = is_experiment
        self.days = days

        if n is None:
            self.n = self.calculate_n_students()
        else:
            self.n = n

        if start is None:
            self.start = pd.Timestamp.now()
        elif isinstance(start, str):
            self.start = pd.to_datetime(start, format="%Y-%m-%d")
        else:
            self.start = start

    def calculate_n_students(self) -> int:
        mu = self.cohort_params["daily_traffic"]["mu"]
        sigma = self.cohort_params["daily_traffic"]["sigma"]
        n = sum(np.random.normal(mu, sigma, self.days).astype(int))
        return n

    def generate_bday(self, years):
        days_offset = np.random.randint(low=0, high=350)
        birthdate = pd.Timestamp.now() - pd.DateOffset(
            years=int(years), days=days_offset
        )
        return birthdate.replace(hour=0, minute=0, second=0, microsecond=0)

    def generate_categorical(self, labels, probabilities, size=1):
        labels = np.array(labels)
        probabilities = np.array(probabilities)
        data = np.random.choice(
            a=labels,
            size=size,
            p=probabilities,
        )
        return data

    def generate_email(self, first: str, last: str) -> str:
        domains = [
            "@yahow.com",
            "@gmall.com",
            "@microsift.com",
            "@hotmeal.com",
        ]
        email = (
            first.lower()
            + "."
            + last.lower()
            + str(np.random.randint(1, 100))
            + np.random.choice(domains, 1)[0]
        )
        return email

    def generate_students(self):
        df = pd.DataFrame()

        # Add columns from demographic params
        for k, v in self.cohort_params["demographics"].items():
            df[k] = self.generate_categorical(
                labels=list(v.keys()),
                probabilities=list(v.values()),
                size=self.n,
            )

        # Create birthday from age
        df["birthday"] = df["age"].apply(self.generate_bday)
        df.drop(columns=["age"], inplace=True)

        # Create name
        df["firstName"] = df["gender"].apply(
            lambda x: names.get_first_name(gender=x)
        )
        df["lastName"] = [names.get_last_name() for x in range(self.n)]

        # Create email
        df["email"] = df[["firstName", "lastName"]].apply(
            lambda x: self.generate_email(*x), axis=1
        )

        # Create timestamp
        stop = self.start + pd.DateOffset(days=self.days)
        df["createdAt"] = [
            randomtimestamp(start=self.start, end=stop) for x in range(self.n)
        ]

        # Reorder columns
        cols = [
            "createdAt",
            "firstName",
            "lastName",
            "email",
            "birthday",
            "gender",
            "highestDegreeEarned",
            "countryISO2",
            "admissionsQuiz",
        ]
        df = df[cols]

        if self.is_experiment:
            df["inExperiment"] = False
            df["group"] = np.nan
            # Get parameters
            response_var = self.cohort_params["experiment"]["response_var"]
            contingency_table = self.cohort_params["experiment"][
                "contingency_table"
            ]

            # Mask to those who haven't taken the quiz
            mask = df["admissionsQuiz"] == "incomplete"
            mask_idx = mask[mask].index.to_list()
            # Create experiment data
            data = []
            for group, response in contingency_table.items():
                for k, v in response.items():
                    count = round(mask.sum() * v)
                    data += [(group, k)] * count

            # Sometimes the lengths of data and mask_idx don't match
            diff = len(data) - len(mask_idx)
            if diff > 0:
                data = data[:-diff]
            if diff < 0:
                mask_idx = mask_idx[:diff]

            # Add to df
            df.loc[mask_idx, ("group", response_var)] = data
            df.loc[mask_idx, "inExperiment"] = True

        return df


class Cohort:
    def __init__(self, **kwargs):
        self._students = None
        self.generator = StudentGenerator(**kwargs)

    def enroll_students(self, confirm_msg=True):
        self._students = self.generator.generate_students()
        if confirm_msg:
            print(f"Added {len(self._students)} students to cohort.")

    def yield_students(self):
        if self._students is None:
            self.enroll_students(confirm_msg=False)
        for s in self._students.itertuples(index=False, name="Student"):
            # Remove fields for subjects not in experiment
            s_dict = s._asdict()
            try:
                if s_dict["inExperiment"] is False:
                    del s_dict["group"]
            except KeyError:
                pass
            yield s_dict

    def get_students(self) -> pd.DataFrame:
        """Getter function"""
        if self._students is None:
            raise AttributeError(
                "You have not enrolled any students in this cohort. Use `enroll_students` first."
            )
        else:
            return self._students



from typing import Optional, Union

import pandas as pd
import pymongo

from teaching_tools.ab_test.data_model import Cohort
from teaching_tools.ab_test.repository import (
    AbstractRepository,
    MongoRepository,
)


class Experiment:
    """Class to generate experimental student data and insert into repository."""

    def __init__(
        self, repo: AbstractRepository, db: str, collection: str
    ) -> None:
        self.attach_repository(repo, db=db)
        self.collection = collection
        self.db = db

    def attach_repository(
        self,
        repository: Union[AbstractRepository, pymongo.MongoClient],
        db: str,
    ) -> None:
        """Attach new repository to experiment.

        Parameters
        ----------
        repository : AbstractRepository

        repo_name : str, optional
            If ``repository`` is ``pymongo.MongoClient``, name of database, by default "wqu-abtest".

        Raises
        ------
        Exception
            Failed to attach repository.
        """
        if isinstance(repository, pymongo.MongoClient):
            self.repo = MongoRepository(repository, db)
        # Pass in Class we build in 7.4
        elif type(repository).__name__ == "MongoRepository":
            client = repository.client
            db = repository.collection._Collection__database._Database__name
            self.repo = MongoRepository(client, db)
        elif isinstance(repository, AbstractRepository):
            self.repo = repository
        elif isinstance(repository, pymongo.collection.Collection):
            db = repository._Collection__database._Database__name
            client = repository._Collection__database._Database__client
            self.repo = MongoRepository(client, db)
        else:
            raise Exception("Cannot attach repository.")

    def add_cohort_to_repository(self, overwrite=False) -> dict:
        """Add cohort of students to repository.

        Parameters
        ----------
        collection : str, optional
            Name of collection to which cohort will be added, by default "ds_applicants"
        overwrite : bool, optional
            Whether to overwrite collection if already exists, by default False
        """
        self.repo.create_collection(
            collection=self.collection, overwrite=overwrite
        )
        r = self.repo.insert(
            self.collection,
            records=self.cohort.yield_students(),
            return_result=True,
        )
        return r

    def run_experiment(self, days: int, start: Optional[str] = None) -> dict:
        """Add experimental student data to repository.

        Parameters
        ----------
        days : int
            Number of days to run the experiment. More days means more students.
        """
        if start is None:
            start = pd.Timestamp.now()
        self.cohort = Cohort(days=days, start=start, is_experiment=True)
        r = self.add_cohort_to_repository()
        return r

    def reset_experiment(self) -> dict:
        """Remove experimental student data from repository.

        Parameters
        ----------
        confirm_msg : bool, optional
            Whether to print confirmation message, by default True
        """
        results = self.repo.drop(
            collection=self.collection,
            query={"inExperiment": {"$exists": True}},
            return_result=True,
        )
        return results



# Inspiration: https://tinyurl.com/2am2r7ga
import abc


class AbstractRepository(abc.ABC):
    @abc.abstractmethod
    def create_collection(self, collection, overwrite=True):
        raise NotImplementedError

    @abc.abstractmethod
    def drop_collection(self, collection):
        raise NotImplementedError

    @abc.abstractmethod
    def find(self, collection, query=None):
        raise NotImplementedError

    @abc.abstractmethod
    def insert(self, collection, records, return_result=False):
        raise NotImplementedError

    @abc.abstractmethod
    def drop(self, collection, query, return_result=False):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, collection, query, new_vals, return_result=False):
        raise NotImplementedError


class MongoRepository(AbstractRepository):
    def __init__(self, client, db):
        self.client = client
        self.db = self.client[db]

    def create_collection(self, collection, overwrite=False):
        collections = self.db.list_collection_names()
        if collection not in collections:
            self.db.create_collection(collection)

        if (collection in collections) and (overwrite):
            self.db[collection].drop()
            self.db.create_collection(collection)

    def drop_collection(self, collection):
        if collection in self.db.list_collection_names():
            self.db[collection].drop()

    def find(self, collection, query=None):
        if query is None:
            q = {}
        else:
            q = query
        return self.db[collection].find(q)

    def insert(self, collection, records, return_result=False):
        r = self.db[collection].insert_many(records)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "inserted_count": len(r.inserted_ids),
            }

    def drop(self, collection, query, return_result=False):
        r = self.db[collection].delete_many(query)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "deleted_count": r.deleted_count,
            }

    def update(self, collection, query, new_vals, return_result=False):
        nv = {"$set": {new_vals}}
        r = self.db[collection].update_many(query, nv)
        if return_result:
            return {
                "acknowledged": r.acknowledged,
                "matched_count": r.matched_count,
                "modified_count": r.modified_count,
            }


class FakeRepository(AbstractRepository):
    def __init__(self, client="client", db="wqu-abtest"):
        self.client = client
        self.db_name = db
        self.db = {}

    def create_collection(self, collection, overwrite=True):
        if (overwrite) or (collection not in self.db.keys()):
            self.db[collection] = list()
        else:
            pass

    def drop_collection(self, collection):
        self.db.pop(collection, None)

    def find(self, collection, query=None):
        if query is None:
            yield self.db[collection]
        else:
            ((key, val),) = query.items()
            for d in self.db[collection]:
                if (key in d) and (d[key] == val):
                    yield d

    def insert(self, collection, records, return_result=False):
        self.db[collection] = self.db[collection] + list(records)
        if return_result:
            return {"acknowledged": True}

    def drop(self, collection: str, query: dict, return_result=False):
        deleted_count = 0
        new_collection = []
        # Need special case for using MongoDB operator for resetting experiment
        if ("inExperiment" in query) and ("$exists" in query["inExperiment"]):
            for d in self.db[collection]:
                if "inExperiment" not in d:
                    new_collection.append(d)
                else:
                    deleted_count += 1
        else:
            ((key, val),) = query.items()
            for d in self.db[collection]:
                if (key not in d) or (d[key] != val):
                    new_collection.append(d)
                else:
                    deleted_count += 1
        self.db[collection] = new_collection
        if return_result:
            return {"acknowledged": True, "deleted_count": deleted_count}

    def update(self, collection, query, new_vals, return_result=False):
        matched_count = 0
        modified_count = 0
        ((key, val),) = query.items()
        ((new_key, new_val),) = new_vals.items()
        for d in self.db[collection]:
            if d[key] == val:
                matched_count += 1
                d[new_key] = new_val
                # I know this counter doesn't exactly work
                modified_count += 1

        if return_result:
            return {
                "acknowledged": True,
                "matched_count": matched_count,
                "modified_count": modified_count,
            }



import itertools
from string import Template

import ipywidgets as widgets
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from IPython.display import Math, display
from matplotlib.patches import Rectangle
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs
from sklearn.exceptions import NotFittedError
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    precision_score,
    recall_score,
)
from sklearn.utils.validation import check_is_fitted


class ConfusionMatrixWidget:
    def __init__(self, model, X_test, y_test):
        self.model = model
        self.X_test = X_test
        self.y_test = y_test
        self._y_pred = None
        self._cnf_matrix = None

    def __repr__(self):
        class_name = type(self).__name__
        model = type(self.model).__name__
        X_test = type(self.X_test).__name__
        y_test = type(self.y_test).__name__
        return "{}(model={}, X_test={}, y_test={})".format(
            class_name, model, X_test, y_test
        )

    def __str__(self):
        class_name = type(self).__name__
        model = type(self.model).__name__
        X_test = type(self.X_test).__name__
        y_test = type(self.y_test).__name__
        return "{}(model={}, X_test={}, y_test={})".format(
            class_name, model, X_test, y_test
        )

    def __generate_predictions(self, thresh=0.5):
        try:
            check_is_fitted(self.model)
        except NotFittedError:
            self.model.fit(self.X_test, self.y_test)

        try:
            pred_proba = self.model.predict_proba(self.X_test)[:, -1]
            self._y_pred = pred_proba > thresh
        except AttributeError:
            raise AttributeError(
                "Widget only works with models that have a `predict_proba` method."
            )

    def __generate_cnf(self, thresh=0.5, rot90=False):
        self.__generate_predictions(thresh)
        _cnf_matrix = confusion_matrix(self.y_test, self._y_pred)
        if rot90:
            _cnf_matrix = np.rot90(_cnf_matrix, axes=(1, 0))
        self._cnf_matrix = _cnf_matrix

    def __acc_score_equation(self, w, short_eq=False):
        with self.accuracy_plt:
            self.accuracy_plt.clear_output(wait=True)
            self.__generate_cnf(self.t_widget.value)
            tn, fp, fn, tp = self._cnf_matrix.ravel()
            score = round(sum([tp + tn]) / sum([tp, fp, fn, tn]), 2)
            if short_eq:
                s = r"""\text{Accuracy} = $SCORE"""
            else:
                s = r"""
                    \text{Accuracy} = \dfrac{
                        \color{orange}{$TP}
                        + \color{blue}{$TN}
                    }{
                        \color{orange}{$TP}
                        + \color{purple}{$FN}
                        + \color{blue}{$TN}
                        + \color{red}{$FP}
                    } = $SCORE
                """
            temp = Template(s).substitute(
                TP=tp, FP=fp, TN=tn, FN=fn, SCORE=score
            )
            display(Math(temp))

    def __precision_score_equation(self, w, short_eq=False):
        with self.precision_plt:
            self.precision_plt.clear_output(wait=True)
            self.__generate_cnf(self.t_widget.value)
            _, fp, _, tp = self._cnf_matrix.ravel()
            score = round(tp / (tp + fp), 2)
            if short_eq:
                s = r"""\text{Precision} = $SCORE"""
            else:
                s = r"""
                    \text{Precision} = \dfrac{
                        \color{orange}{$TP}
                    }{
                        \color{orange}{$TP}
                        + \color{red}{$FP}
                    } = $SCORE
                """
            temp = Template(s).substitute(TP=tp, FP=fp, SCORE=score)
            display(Math(temp))

    def __recall_score_equation(self, w, short_eq=False):
        with self.recall_plt:
            self.recall_plt.clear_output(wait=True)
            self.__generate_cnf(self.t_widget.value)
            _, _, fn, tp = self._cnf_matrix.ravel()
            score = round(tp / (tp + fn), 2)
            if short_eq:
                s = r"""\text{Recall} = $SCORE"""
            else:
                s = r"""
                    \text{Recall} = \dfrac{
                        \color{orange}{$TP}
                    }{
                        \color{orange}{$TP}
                        + \color{purple}{$FN}
                    } = $SCORE
                """
            temp = Template(s).substitute(TP=tp, FN=fn, SCORE=score)
            display(Math(temp))

    def __three_short(self, w):
        with self.three_short_plt:
            self.three_short_plt.clear_output(wait=True)
            self.__generate_cnf(self.t_widget.value)
            prec = round(
                precision_score(self.y_test, self._y_pred, zero_division=0), 2
            )
            acc = round(accuracy_score(self.y_test, self._y_pred), 2)
            rec = round(
                recall_score(self.y_test, self._y_pred, zero_division=0), 2
            )

            s = r"""
                \begin{align}
                    \text{Accuracy} & = $ACC\\
                    \text{Precision} & = $PREC\\
                    \text{Recall} & = $REC
                \end{align}
            """

            temp = Template(s).substitute(ACC=acc, PREC=prec, REC=rec)
            display(Math(temp))

    def __costs_equation(
        self, w, costs={"tn": 0, "fp": 500, "fn": 50000, "tp": 0}
    ):
        with self.costs_plt:
            self.costs_plt.clear_output(wait=True)
            self.__generate_cnf(self.t_widget.value)
            _, fp, fn, _ = self._cnf_matrix.ravel()
            fp_total = format(fp * costs["fp"], "6,d").replace(",", "{,}")
            fn_total = format(fn * costs["fn"], "6,d").replace(",", "{,}")
            grand_total = format(
                (fn * costs["fn"] + fp * costs["fp"]), "6,d"
            ).replace(",", "{,}")
            s = r"""
                    \begin{align}
                        \text{Wasted admin costs:} &~€ & $FP_TOTAL\\
                        \text{Wasted legal costs:} &~€ & $FN_TOTAL \\
                        \hline \\
                        \text{Total wasted costs:} &~€ & $GRAND_TOTAL
                    \end{align}
            """

            temp = Template(s).substitute(
                FP_TOTAL=fp_total, FN_TOTAL=fn_total, GRAND_TOTAL=grand_total
            )
            display(Math(temp))

    def __cnf_matrix(self, w, font_size=25):
        # https://github.com/jupyter-widgets/ipywidgets/issues/1919
        with self.cnf_plot:
            # Clear output
            self.cnf_plot.clear_output(wait=True)

            # Create confusion matrix
            self.__generate_cnf(self.t_widget.value, rot90=True)

            # Set colors
            color_matrix = np.array([["purple", "blue"], ["orange", "red"]])

            # Create fig
            fig = plt.figure(figsize=(10, 10))  # noQA F841
            ax = plt.subplot(111, aspect="equal")

            # Create plot
            for i, j in itertools.product(
                range(self._cnf_matrix.shape[0]),
                range(self._cnf_matrix.shape[1]),
            ):
                # Add patch
                sq = Rectangle(
                    xy=(i, j),
                    width=1,
                    height=1,
                    fc=color_matrix[i, j],
                    alpha=0.5,
                    ec="black",
                    label="Label",
                )
                ax.add_patch(sq)

                # Calculate label position
                rx, ry = sq.get_xy()
                cx = rx + sq.get_width() / 2.0
                cy = ry + sq.get_height() / 2.0

                # Add labels
                ax.annotate(
                    self._cnf_matrix[i, j],
                    (cx, cy),
                    color="black",
                    ha="center",
                    va="center",
                    fontsize=font_size,
                )
            ax.autoscale_view()
            ax.set_frame_on(False)
            ax.yaxis.label.set_size(font_size)
            ax.xaxis.label.set_size(font_size)
            plt.xticks(
                ticks=(0.5, 1.5), labels=[False, True], fontsize=font_size
            )
            plt.yticks(
                ticks=(0.5, 1.5), labels=[True, False], fontsize=font_size
            )
            plt.xlabel("Predicted label")
            plt.ylabel("True label")
            plt.show()

    def show(self):
        # Create widgets
        self.cnf_plot = widgets.Output(
            layout=widgets.Layout(height="300px", width="300px")
        )
        self.accuracy_plt = widgets.Output(
            layout=widgets.Layout(height="80px", width="400px")
        )
        self.precision_plt = widgets.Output(
            layout=widgets.Layout(height="80px", width="400px")
        )
        self.recall_plt = widgets.Output(
            layout=widgets.Layout(height="80px", width="400px")
        )
        self.t_widget = widgets.FloatSlider(
            value=0.5,
            min=0.0,
            max=1,
            step=0.1,
            description="Threshold:",
            continuous_update=False,
        )

        # Run helper funtions
        self.__cnf_matrix([])
        self.__acc_score_equation([])
        self.__precision_score_equation([])
        self.__recall_score_equation([])

        # Attach threshold widget to helper functions
        self.t_widget.observe(self.__cnf_matrix)
        self.t_widget.observe(self.__acc_score_equation)
        self.t_widget.observe(self.__precision_score_equation)
        self.t_widget.observe(self.__recall_score_equation)

        equations = widgets.VBox(
            [self.accuracy_plt, self.precision_plt, self.recall_plt]
        )
        hbx = widgets.HBox([self.cnf_plot, equations])
        display(self.t_widget, hbx)

    def show_eu(self, costs={"tn": 0, "fp": 500, "fn": 50000, "tp": 0}):
        try:
            costs = dict(costs)
        except ValueError:
            raise ValueError("`costs` should by a dictionary.")

        # Create widgets
        self.cnf_plot = widgets.Output(
            layout=widgets.Layout(height="300px", width="300px")
        )
        self.three_short_plt = widgets.Output(
            layout=widgets.Layout(height="150px", width="150px")
        )
        self.costs_plt = widgets.Output(
            layout=widgets.Layout(height="150px", width="300px")
        )
        self.t_widget = widgets.FloatSlider(
            value=0.5,
            min=0.0,
            max=1,
            step=0.1,
            description="Threshold:",
            continuous_update=False,
        )

        # Run helper funtions
        self.__cnf_matrix([])
        self.__three_short([])
        self.__costs_equation([])

        # Attach threshold widget to helper functions
        self.t_widget.observe(self.__cnf_matrix)
        self.t_widget.observe(self.__three_short)
        self.t_widget.observe(self.__costs_equation)

        vbx = widgets.VBox([self.three_short_plt, self.costs_plt])
        hbx = widgets.HBox([self.cnf_plot, vbx])
        display(self.t_widget, hbx)


class ClusterWidget:
    def __init__(
        self, X=None, n_clusters=3, seeds=np.array([[0, 4], [4, 2], [-2, 4]])
    ):
        self.n_clusters = n_clusters
        self.seeds = seeds
        if X is None:
            self.X = self.__make_cluster_data()
        else:
            self.X = X

    def __make_cluster_data(self):
        X, _ = make_blobs(
            n_samples=300,
            n_features=2,
            centers=3,
            cluster_std=2,
            random_state=0,
        )
        return X

    def __plot_kmeans_steps(self, w):
        with self.scatter_plot:
            step = self.step_widget.value
            self.scatter_plot.clear_output(wait=True)

            iters = step // 2
            if iters:
                kmeans = KMeans(
                    n_clusters=self.n_clusters,
                    max_iter=iters,
                    n_init=1,
                    init=self.seeds,
                )
                kmeans.fit(self.X)
                centroids = kmeans.cluster_centers_
                labels = kmeans.labels_
            else:
                centroids = self.seeds
                labels = "0.5"
            if step % 2:
                kmeans = KMeans(
                    n_clusters=self.n_clusters,
                    max_iter=iters + 1,
                    n_init=1,
                    init=self.seeds,
                )
                kmeans.fit(self.X)
                labels = kmeans.labels_

            title = (
                f"Step {step+0}: "
                + ["Set Centroids", "Assign Clusters"][step % 2]
            )
            if step > 1:
                title += " (again)"

            plt.scatter(*self.X.T, c=labels, cmap="viridis", alpha=0.5)
            plt.scatter(
                *centroids.T,
                c=range(self.n_clusters),
                cmap="viridis",
                marker="*",
                s=150,
                linewidths=1,
                edgecolors="k",
            )
            plt.title(title)
            plt.show()

    def show(self):
        # Create widgets
        self.scatter_plot = widgets.Output(
            layout=widgets.Layout(height="400px", width="400px")
        )
        self.step_widget = widgets.IntSlider(
            min=0,
            max=10,
            value=0,
            step=1,
            description="Step:",
            continuous_update=False,
        )

        # Run helper functions
        self.__plot_kmeans_steps([])

        # Attach widget
        self.step_widget.observe(self.__plot_kmeans_steps)

        # Layout
        vbx = widgets.VBox([self.step_widget, self.scatter_plot])

        # Display
        display(vbx)


class SCFClusterWidget:
    def __init__(self, x, y, n_clusters=3):
        self.x = pd.Series(x)
        self.y = pd.Series(y)
        self.n_clusters = n_clusters
        self.__set_seeds()

    def __repr__(self):
        class_name = type(self).__name__
        return f"{class_name}(x={self.x.shape}, y={self.y.shape}, n_clusters={self.n_clusters})"

    def __set_seeds(self):
        x_coords = np.linspace(
            self.x.min(), self.x.max(), self.n_clusters + 2
        )[1:-1].reshape(-1, 1)
        y_coords = np.linspace(
            self.y.min(), self.y.max(), self.n_clusters + 2
        )[1:-1].reshape(-1, 1)
        self.seeds = np.hstack([x_coords, y_coords])

    def make_model(self, step):
        iters = step // 2

        # Step 0
        if not iters:
            self.centroids = self.seeds
            self.labels = "0.5"

        # Step 2 and following evens
        if iters:
            self.kmeans = KMeans(
                n_clusters=self.n_clusters,
                max_iter=iters,
                n_init=1,
                init=self.seeds,
            )
            self.kmeans.fit(pd.DataFrame({"x": self.x, "y": self.y}))
            self.centroids = self.kmeans.cluster_centers_
            self.labels = self.kmeans.labels_

        # Step 1 and following odds
        if step % 2:
            self.kmeans = KMeans(
                n_clusters=self.n_clusters,
                max_iter=iters + 1,
                n_init=1,
                init=self.seeds,
            )
            self.kmeans.fit(pd.DataFrame({"x": self.x, "y": self.y}))
            self.labels = self.kmeans.labels_

    def __make_plot(self, w):
        with self.plot:
            plt.rcParams.update({"font.size": 25})
            self.plot.clear_output(wait=True)
            step = self.slider.value
            self.make_model(step)

            # Create fig
            fig, (ax1, ax2) = plt.subplots(
                nrows=1, ncols=2, figsize=(19, 7), constrained_layout=True
            )

            # Left axis, data
            ax1.scatter(
                self.x / 1e6,
                self.y / 1e6,
                c=self.labels,
                cmap="Accent",
                s=200,
                alpha=0.25,
            )

            # Left axis, centroids
            ax1.scatter(
                *(self.centroids / 1e6).T,
                c=range(self.n_clusters),
                cmap="Accent",
                alpha=1,
                marker="*",
                s=600,
                linewidths=1,
                edgecolors="black",
            )

            ax1.set_xlabel("Household Debt [$1M]")
            ax1.set_ylabel("Home Value [$1M]")

            # Right axis, log transform
            np.seterr(divide="ignore")
            log_x = np.log(self.x)
            log_y = np.log(self.y)
            log_centroids = np.log(self.centroids)
            np.seterr(divide="warn")

            # Right axis, data
            ax2.scatter(
                x=log_x,
                y=log_y,
                c=self.labels,
                cmap="Accent",
                alpha=0.25,
                s=200,
            )

            # Right axis, centroids
            ax2.scatter(
                *log_centroids.T,
                c=range(self.n_clusters),
                cmap="Accent",
                alpha=1,
                marker="*",
                s=600,
                linewidths=1,
                edgecolors="black",
            )

            ax2.set_xlabel("Household Debt [log($)]")
            ax2.set_ylabel("Home Value [log($)]")

            # Plot title
            title = (
                f"Step {step+0}: "
                + ["Set Centroids", "Assign Clusters"][step % 2]
            )
            if step > 1:
                title += " (again)"
            plt.suptitle(title)

            plt.show()
            matplotlib.rcdefaults()

    def show(self):
        # Create plot widget
        self.plot = widgets.Output(
            layout=widgets.Layout(height="220px", width="550px")
        )

        # Create slider widget
        self.slider = widgets.IntSlider(
            min=0,
            max=10,
            value=0,
            step=1,
            description="Step:",
            continuous_update=False,  # Always use
        )

        # Initialize plot function
        # Note empty list for `w`
        self.__make_plot([])

        # Attach slider to plot function
        self.slider.observe(self.__make_plot)

        # Set layout (vertical box)
        vbx = widgets.VBox([self.slider, self.plot])

        # Display widget
        display(vbx)


class ExampleWidget:
    """So that I don't forget how to do this."""

    def __init__(self):
        """Stuff you want from the user"""
        pass

    def __plot_datapoint(self, w):
        """Plots a single datapoint

        Parameters
        ----------
        w : list
            Dummy variable that prevents crashing. Don't use.
        """
        # `plot` defined below
        with self.plot:
            # Clear plot
            self.plot.clear_output(wait=True)
            # Make plot
            # If you need val from `slider`, defined below
            val = self.slider.value
            plt.scatter(val, val, s=200)
            plt.xlim((0, 6))
            plt.ylim((0, 6))
            plt.title("Hot to Plot")
            plt.show()

    def show(self):
        # Create plot widget
        self.plot = widgets.Output(
            layout=widgets.Layout(height="400px", width="400px")
        )

        # Crete slider widget
        self.slider = widgets.IntSlider(
            min=1,
            max=5,
            value=2,
            step=1,
            description="Val",
            continuous_update=False,  # Always use
        )

        # Initialize plot function
        # Note empty list for `w`
        self.__plot_datapoint([])

        # Attach slider to plot function
        self.slider.observe(self.__plot_datapoint)

        # Set layout (vertical box)
        vbx = widgets.VBox([self.slider, self.plot])

        # Display widget
        display(vbx)
